==> kotlin-compiler-1.4.0-rc-sources.jar: org/jetbrains/kotlin/ir/util/IrFakeOverrideUtils.kt <==
/*
 * Copyright 2010-2019 JetBrains s.r.o. and Kotlin Programming Language contributors.
 * Use of this source code is governed by the Apache 2.0 license that can be found in the license/LICENSE.txt file.
 */

package org.jetbrains.kotlin.ir.util

import org.jetbrains.kotlin.descriptors.Modality
import org.jetbrains.kotlin.ir.declarations.*
import org.jetbrains.kotlin.utils.addToStdlib.safeAs

val IrDeclaration.isReal: Boolean get() = !isFakeOverride

val IrDeclaration.isFakeOverride: Boolean
    get() = when (this) {
        is IrSimpleFunction -> isFakeOverride
        is IrProperty -> isFakeOverride
        is IrField -> isFakeOverride
        else -> false
    }

val IrSimpleFunction.target: IrSimpleFunction
    get() = if (modality == Modality.ABSTRACT)
        this
    else
        resolveFakeOverride() ?: error("Could not resolveFakeOverride() for ${this.render()}")

val IrFunction.target: IrFunction get() = when (this) {
    is IrSimpleFunction -> this.target
    is IrConstructor -> this
    else -> error(this)
}

fun IrField.resolveFakeOverride(): IrField? {
    var toVisit = setOf(this)
    val nonOverridden = mutableSetOf<IrField>()
    while (toVisit.isNotEmpty()) {
        nonOverridden += toVisit.filter { it.overriddenSymbols.isEmpty() }
        toVisit = toVisit.flatMap { it.overriddenSymbols }.map { it.owner }.toSet()
    }
    return nonOverridden.singleOrNull()
}
fun IrSimpleFunction.collectRealOverrides(toSkip: (IrSimpleFunction) -> Boolean = { false }): Set<IrSimpleFunction> {
    if (isReal && !toSkip(this)) return setOf(this)

    val visited = mutableSetOf<IrSimpleFunction>()
    val realOverrides = mutableSetOf<IrSimpleFunction>()

    fun collectRealOverrides(func: IrSimpleFunction) {
        if (!visited.add(func)) return

        if (func.isReal && !toSkip(func)) {
            realOverrides += func
        } else {
            func.overriddenSymbols.forEach { collectRealOverrides(it.owner) }
        }
    }

    overriddenSymbols.forEach { collectRealOverrides(it.owner) }

    fun excludeRepeated(func: IrSimpleFunction) {
        if (!visited.add(func)) return

        func.overriddenSymbols.forEach {
            realOverrides.remove(it.owner)
            excludeRepeated(it.owner)
        }
    }

    visited.clear()
    realOverrides.toList().forEach { excludeRepeated(it) }

    return realOverrides
}

// TODO: use this implementation instead of any other
fun IrSimpleFunction.resolveFakeOverride(allowAbstract: Boolean = false, toSkip: (IrSimpleFunction) -> Boolean = { false }): IrSimpleFunction? {
    val reals = collectRealOverrides(toSkip)
    return if (allowAbstract) {
        if (reals.isEmpty()) error("No real overrides for ${this.render()}")
        reals.first()
    } else {
        reals
            .filter { it.modality != Modality.ABSTRACT }
            .let { realOverrides ->
                // Kotlin forbids conflicts between overrides, but they may trickle down from Java.
                realOverrides.singleOrNull { it.parent.safeAs<IrClass>()?.isInterface != true }
                // TODO: We take firstOrNull instead of singleOrNull here because of KT-36188.
                    ?: realOverrides.firstOrNull()
            }
    }
}

==> kotlin-compiler-1.4.21-sources.jar: org/jetbrains/kotlin/ir/util/IrFakeOverrideUtils.kt <==
/*
 * Copyright 2010-2019 JetBrains s.r.o. and Kotlin Programming Language contributors.
 * Use of this source code is governed by the Apache 2.0 license that can be found in the license/LICENSE.txt file.
 */

package org.jetbrains.kotlin.ir.util

import org.jetbrains.kotlin.descriptors.Modality
import org.jetbrains.kotlin.ir.declarations.*
import org.jetbrains.kotlin.utils.addToStdlib.safeAs

val IrDeclaration.isReal: Boolean get() = !isFakeOverride

val IrDeclaration.isFakeOverride: Boolean
    get() = when (this) {
        is IrSimpleFunction -> isFakeOverride
        is IrProperty -> isFakeOverride
        else -> false
    }

val IrSimpleFunction.target: IrSimpleFunction
    get() = if (modality == Modality.ABSTRACT)
        this
    else
        resolveFakeOverride() ?: error("Could not resolveFakeOverride() for ${this.render()}")

val IrFunction.target: IrFunction get() = when (this) {
    is IrSimpleFunction -> this.target
    is IrConstructor -> this
    else -> error(this)
}

fun IrSimpleFunction.collectRealOverrides(toSkip: (IrSimpleFunction) -> Boolean = { false }): Set<IrSimpleFunction> {
    if (isReal && !toSkip(this)) return setOf(this)

    return this.overriddenSymbols
        .map { it.owner }
        .collectAndFilterRealOverrides {
            require(it is IrSimpleFunction) { "Expected IrSimpleFunction: ${it.render()}" }
            toSkip(it)
        }
        .map { it as IrSimpleFunction }
        .toSet()
}

fun Collection<IrOverridableMember>.collectAndFilterRealOverrides(toSkip: (IrOverridableMember) -> Boolean = { false }): Set<IrOverridableMember> {

    val visited = mutableSetOf<IrOverridableMember>()
    val realOverrides = mutableSetOf<IrOverridableMember>()

    fun overriddenSymbols(declaration: IrOverridableMember) = when (declaration) {
        is IrSimpleFunction -> declaration.overriddenSymbols
        is IrProperty -> (declaration.getter ?: declaration.setter)
            ?.overriddenSymbols?.mapNotNull { it.owner.correspondingPropertySymbol }
            ?: emptyList()
        else -> error("Unexpected overridable member: ${declaration.render()}")
    }

    fun collectRealOverrides(member: IrOverridableMember) {
        if (!visited.add(member)) return

        if (member.isReal && !toSkip(member)) {
            realOverrides += member
        } else {
            overriddenSymbols(member).forEach { collectRealOverrides(it.owner as IrOverridableMember) }
        }
    }

    this.forEach { collectRealOverrides(it) }

    fun excludeRepeated(member: IrOverridableMember) {
        if (!visited.add(member)) return

        overriddenSymbols(member).forEach {
            realOverrides.remove(it.owner)
            excludeRepeated(it.owner as IrOverridableMember)
        }
    }

    visited.clear()
    realOverrides.toList().forEach { excludeRepeated(it) }

    return realOverrides
}

// TODO: use this implementation instead of any other
fun IrSimpleFunction.resolveFakeOverride(allowAbstract: Boolean = false, toSkip: (IrSimpleFunction) -> Boolean = { false }): IrSimpleFunction? {
    val reals = collectRealOverrides(toSkip)
    return if (allowAbstract) {
        if (reals.isEmpty()) error("No real overrides for ${this.render()}")
        reals.first()
    } else {
        reals
            .filter { it.modality != Modality.ABSTRACT }
            .let { realOverrides ->
                // Kotlin forbids conflicts between overrides, but they may trickle down from Java.
                realOverrides.singleOrNull { it.parent.safeAs<IrClass>()?.isInterface != true }
                // TODO: We take firstOrNull instead of singleOrNull here because of KT-36188.
                    ?: realOverrides.firstOrNull()
            }
    }
}
